function [whitesigs,whitemat,dewhitemat,avgsigs,sigpcov]=whitening(signals,pcafac,filttype)
%
% Spatial whitening for possibly complex-valued signals
%   - only diagonalze the covariance matrix E[x*x']
%   - pseudo-covariance E[x*x.'] is not diagnalized
%   - pcafac: 0<p<1,cumcont; 1,2,... number; []: full (=rank)
%             0: symmetric whitening (no dim reduction)
%

% Last update: 15-Sep-2015
% Author: Jun-ichiro Hirayama 

if nargin<2, pcafac = []; end
if nargin<3, filttype = 'complex'; end
%Centering
avgsigs = mean(signals,2);
signals = bsxfun(@minus, signals,avgsigs);
%Spatial whitening with dimensionality reduction
n = size(signals,2);
sigcov  = (signals * signals')./n;
if nargout>4
    sigpcov = (signals * signals.')./n;
end
if strcmp(filttype,'real'), sigcov = real(sigcov); end
[eigvecs,eigvals] = eig(sigcov);
eigvals = diag(eigvals);
eigvals(eigvals<0) = 0; %just in case
[eigvals,order]=sort(eigvals,'descend');
eigvecs = eigvecs(:,order);
%Determine the dimensionality
if isempty(pcafac) || pcafac == 0
    d = sum(eigvals>numel(eigvals).*eps(max(eigvals)));
elseif pcafac>=1
    d = pcafac;
elseif pcafac>0
    d = find(cumsum(eigvals)./sum(eigvals) >= pcafac,1);
end
scales = sqrt(eigvals(1:d));
eigvecs = eigvecs(:,1:d);
if pcafac==0 %symmetric whitening: (cov(X)^(1/2))*X
    whitemat = eigvecs*diag(1./scales)*eigvecs';
    dewhitemat = eigvecs*diag(scales)*eigvecs';
else         %PCA-based whitening
    whitemat = diag(1./scales)*eigvecs';
    dewhitemat = eigvecs*diag(scales);
end
whitesigs = whitemat * signals;

return
